from typing import cast

import gym
import torch
from torch import nn

from extensions.rl_lighthouse.lighthouse_models import LinearAdvisorActorCritic
from extensions.rl_minigrid.minigrid_experiments.base import (
    MiniGridBaseExperimentConfig,
)
from extensions.rl_minigrid.minigrid_models import MiniGridSimpleConvRNN
from extensions.rl_minigrid.minigrid_sensors import EgocentricMiniGridSensor
from onpolicy_sync.losses.advisor import (
    AdvisorImitationStage,
    AdvisorWeightedStage,
)
from rl_base.sensor import SensorSuite
from utils.experiment_utils import Builder, PipelineStage, LinearDecay


class MiniGridDaggerThenAdvisorFixedAlphaDiffHeadsExperimentConfig(
    MiniGridBaseExperimentConfig
):
    """Training with DAgger followed adaptive reweighing."""

    GPU_ID = 1 if torch.cuda.is_available() else None
    USE_EXPERT = True
    SAME_WEIGHTS_FOR_ADVISOR_HEAD = False

    @classmethod
    def extra_tag(cls):
        return "DaggerThenAdvisorDiffHeadsFixedAlpha__alpha_{}__lr_{}__tf_{}".format(
            cls.alpha(), cls.lr(), cls.tf_ratio(),
        )

    @classmethod
    def training_pipeline(cls, **kwargs):
        training_steps = cls.TOTAL_TRAIN_STEPS
        steps_advisor_warmup_stage = int(training_steps * cls.tf_ratio())
        steps_advisor_weighted_stage = training_steps - steps_advisor_warmup_stage

        ppo_info = cls.rl_loss_default("ppo", steps=steps_advisor_weighted_stage)
        fixed_alpha = cls.alpha()
        return cls._training_pipeline(
            named_losses={
                "advisor_imitation_warmup": Builder(AdvisorImitationStage,),
                "advisor_loss": Builder(
                    AdvisorWeightedStage,
                    kwargs={"rl_loss": ppo_info["loss"], "fixed_alpha": fixed_alpha,},
                ),
            },
            pipeline_stages=[
                PipelineStage(
                    loss_names=["advisor_imitation_warmup"],
                    max_stage_steps=steps_advisor_warmup_stage,
                    teacher_forcing=LinearDecay(
                        startp=1.0, endp=0.0, steps=steps_advisor_warmup_stage,
                    ),
                ),
                PipelineStage(
                    loss_names=["advisor_loss"],
                    max_stage_steps=steps_advisor_weighted_stage,
                    early_stopping_criterion=cls.task_info().get(
                        "early_stopping_criterion"
                    ),
                ),
            ],
            num_mini_batch=ppo_info["num_mini_batch"],
            update_repeats=ppo_info["update_repeats"],
        )

    @classmethod
    def create_model(cls, **kwargs) -> nn.Module:
        sensors = cls.get_sensors()
        return MiniGridSimpleConvRNN(
            action_space=gym.spaces.Discrete(
                len(cls.task_info()["task_class"].class_action_names())
            ),
            num_objects=cast(EgocentricMiniGridSensor, sensors[0]).num_objects,
            num_colors=cast(EgocentricMiniGridSensor, sensors[0]).num_colors,
            num_states=cast(EgocentricMiniGridSensor, sensors[0]).num_states,
            observation_space=SensorSuite(sensors).observation_spaces,
            hidden_size=128,
            rnn_type=cls.RNN_TYPE,
            head_type=Builder(  # type: ignore
                LinearAdvisorActorCritic,
                kwargs={"ensure_same_weights": cls.SAME_WEIGHTS_FOR_ADVISOR_HEAD},
            ),
        )
